import codecs
import dataclasses
import linecache
import os
import tempfile
from ast import literal_eval
from collections import ChainMap
from contextlib import contextmanager
from itertools import chain, islice
from types import CodeType
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Union,
    Set,
    Type,
    NoReturn,
)

from typing_extensions import Protocol

import jinja2
import jinja2.ext
import jinja2.nativetypes
import jinja2.nodes
import jinja2.parser
import jinja2.sandbox

from dbt_common.tests import test_caching_enabled
from dbt_common.utils.jinja import (
    get_dbt_macro_name,
    get_docs_macro_name,
    get_materialization_macro_name,
    get_test_macro_name,
)
from dbt_common.clients._jinja_blocks import (
    BlockIterator,
    BlockData,
    BlockTag,
    TagIterator,
    ExtractWarning,
)

from dbt_common.exceptions import (
    CompilationError,
    DbtInternalError,
    CaughtMacroErrorWithNodeError,
    MaterializationArgError,
    JinjaRenderingError,
    UndefinedCompilationError,
    DbtRuntimeError,
)
from dbt_common.exceptions.macros import MacroReturn, UndefinedMacroError, CaughtMacroError


SUPPORTED_LANG_ARG = jinja2.nodes.Name("supported_languages", "param")

# Global which can be set by dependents of dbt-common (e.g. core via flag parsing)
MACRO_DEBUGGING: Union[str, bool] = False

_ParseReturn = Union[jinja2.nodes.Node, List[jinja2.nodes.Node]]


# Temporary type capturing the concept the functions in this file expect for a "node"
class _NodeProtocol(Protocol):
    pass


def _linecache_inject(source: str, write: bool) -> str:
    if write:
        # this is the only reliable way to accomplish this. Obviously, it's
        # really darn noisy and will fill your temporary directory
        tmp_file = tempfile.NamedTemporaryFile(
            prefix="dbt-macro-compiled-",
            suffix=".py",
            delete=False,
            mode="w+",
            encoding="utf-8",
        )
        tmp_file.write(source)
        filename = tmp_file.name
    else:
        # `codecs.encode` actually takes a `bytes` as the first argument if
        # the second argument is 'hex' - mypy does not know this.
        rnd = codecs.encode(os.urandom(12), "hex")
        filename = rnd.decode("ascii")

    # put ourselves in the cache
    cache_entry = (len(source), None, [line + "\n" for line in source.splitlines()], filename)
    # linecache does in fact have an attribute `cache`, thanks
    linecache.cache[filename] = cache_entry
    return filename


@dataclasses.dataclass
class MacroType:
    name: str
    type_params: List["MacroType"] = dataclasses.field(default_factory=list)


class MacroFuzzParser(jinja2.parser.Parser):
    def parse_macro(self) -> jinja2.nodes.Macro:
        node = jinja2.nodes.Macro(lineno=next(self.stream).lineno)

        # modified to fuzz macros defined in the same file. this way
        # dbt can understand the stack of macros being called.
        #  - @cmcarthur
        node.name = get_dbt_macro_name(self.parse_assign_target(name_only=True).name)

        self.parse_signature(node)
        node.body = self.parse_statements(("name:endmacro",), drop_needle=True)
        return node

    def parse_signature(self, node: Union[jinja2.nodes.Macro, jinja2.nodes.CallBlock]) -> None:
        """Overrides the default jinja Parser.parse_signature method, modifying
        the original implementation to allow macros to have typed parameters."""

        # Jinja does not support extending its node types, such as Macro, so
        # at least while typed macros are experimental, we will patch the
        # information onto the existing types.
        setattr(node, "arg_types", [])
        setattr(node, "has_type_annotations", False)

        args = node.args = []  # type: ignore
        defaults = node.defaults = []  # type: ignore

        self.stream.expect("lparen")
        while self.stream.current.type != "rparen":
            if args:
                self.stream.expect("comma")

            arg = self.parse_assign_target(name_only=True)
            arg.set_ctx("param")

            type_name: Optional[str]
            if self.stream.skip_if("colon"):
                node.has_type_annotations = True  # type: ignore
                type_name = self.parse_type_name()
            else:
                type_name = ""

            node.arg_types.append(type_name)  # type: ignore

            if self.stream.skip_if("assign"):
                defaults.append(self.parse_expression())
            elif defaults:
                self.fail("non-default argument follows default argument")

            args.append(arg)
        self.stream.expect("rparen")

    def parse_type_name(self) -> MacroType:
        # NOTE: Types syntax is validated here, but not whether type names
        # are valid or have correct parameters.

        # A type name should consist of a name (i.e. 'Dict')...
        type_name = self.stream.expect("name").value
        type = MacroType(type_name)

        # ..and an optional comma-delimited list of type parameters
        # as in the type declaration 'Dict[str, str]'
        if self.stream.skip_if("lbracket"):
            while self.stream.current.type != "rbracket":
                if type.type_params:
                    self.stream.expect("comma")
                param_type = self.parse_type_name()
                type.type_params.append(param_type)

            self.stream.expect("rbracket")

        return type


class MacroFuzzEnvironment(jinja2.sandbox.SandboxedEnvironment):
    def _parse(
        self, source: str, name: Optional[str], filename: Optional[str]
    ) -> jinja2.nodes.Template:
        return MacroFuzzParser(self, source, name, filename).parse()

    def _compile(self, source: str, filename: str) -> CodeType:
        """
        Override jinja's compilation. Use to stash the rendered source inside
        the python linecache for debugging when the appropriate environment
        variable is set.

        If the value is 'write', also write the files to disk.
        WARNING: This can write a ton of data if you aren't careful.
        """
        if filename == "<template>" and MACRO_DEBUGGING:
            write = MACRO_DEBUGGING == "write"
            filename = _linecache_inject(source, write)

        return super()._compile(source, filename)  # type: ignore


class MacroFuzzTemplate(jinja2.nativetypes.NativeTemplate):
    environment_class = MacroFuzzEnvironment  # type: ignore

    def new_context(
        self,
        vars: Optional[Dict[str, Any]] = None,
        shared: bool = False,
        locals: Optional[Mapping[str, Any]] = None,
    ) -> jinja2.runtime.Context:
        # This custom override makes the assumption that the locals and shared
        # parameters are not used, so enforce that.
        if shared or locals:
            raise Exception(
                "The MacroFuzzTemplate.new_context() override cannot use the "
                "shared or locals parameters."
            )

        vars = {} if vars is None else vars
        parent = ChainMap(vars, self.globals) if self.globals else vars

        return self.environment.context_class(self.environment, parent, self.name, self.blocks)

    def render(self, *args: Any, **kwargs: Any) -> Any:
        if kwargs or len(args) != 1:
            raise Exception(
                "The MacroFuzzTemplate.render() override requires exactly one argument."
            )

        ctx = self.new_context(args[0])

        try:
            return self.environment_class.concat(  # type: ignore
                self.root_render_func(ctx)  # type: ignore
            )
        except Exception:
            return self.environment.handle_exception()


MacroFuzzEnvironment.template_class = MacroFuzzTemplate


class NativeSandboxEnvironment(MacroFuzzEnvironment):
    code_generator_class = jinja2.nativetypes.NativeCodeGenerator


class TextMarker(str):
    """A special native-env marker that indicates a value is text and is not to be evaluated.

    Use this to prevent your numbery-strings from becoming numbers!
    """


class NativeMarker(str):
    """A special native-env marker that indicates the field should be passed to literal_eval."""


class BoolMarker(NativeMarker):
    pass


class NumberMarker(NativeMarker):
    pass


def _is_number(value: Any) -> bool:
    return isinstance(value, (int, float)) and not isinstance(value, bool)


def quoted_native_concat(nodes: Iterator[str]) -> Any:
    """Handle special case for native_concat from the NativeTemplate.

    This is almost native_concat from the NativeTemplate, except in the
    special case of a single argument that is a quoted string and returns a
    string, the quotes are re-inserted.
    """
    head = list(islice(nodes, 2))

    if not head:
        return ""

    if len(head) == 1:
        raw = head[0]
        if isinstance(raw, TextMarker):
            return str(raw)
        elif not isinstance(raw, NativeMarker):
            # return non-strings as-is
            return raw
    else:
        # multiple nodes become a string.
        return "".join([str(v) for v in chain(head, nodes)])

    try:
        result = literal_eval(raw)
    except (ValueError, SyntaxError, MemoryError):
        result = raw
    if isinstance(raw, BoolMarker) and not isinstance(result, bool):
        raise JinjaRenderingError(f"Could not convert value '{raw!s}' into type 'bool'")
    if isinstance(raw, NumberMarker) and not _is_number(result):
        raise JinjaRenderingError(f"Could not convert value '{raw!s}' into type 'number'")

    return result


class NativeSandboxTemplate(jinja2.nativetypes.NativeTemplate):  # mypy: ignore
    environment_class = NativeSandboxEnvironment  # type: ignore

    def render(self, *args: Any, **kwargs: Any) -> Any:
        """Render the template to produce a native Python type.

        If the result is a single node, its value is returned. Otherwise,
        the nodes are concatenated as strings. If the result can be parsed
        with :func:`ast.literal_eval`, the parsed value is returned.
        Otherwise, the string is returned.
        """
        vars = args[0]

        try:
            return quoted_native_concat(self.root_render_func(self.new_context(vars)))
        except Exception:
            return self.environment.handle_exception()


class MacroProtocol(Protocol):
    name: str
    macro_sql: str


NativeSandboxEnvironment.template_class = NativeSandboxTemplate  # type: ignore


class TemplateCache:
    def __init__(self) -> None:
        self.file_cache: Dict[str, jinja2.Template] = {}

    def get_node_template(self, node: MacroProtocol) -> jinja2.Template:
        key = node.macro_sql

        if key in self.file_cache:
            return self.file_cache[key]

        template = get_template(
            string=node.macro_sql,
            ctx={},
            node=node,
        )

        self.file_cache[key] = template
        return template

    def clear(self) -> None:
        self.file_cache.clear()


template_cache = TemplateCache()


class BaseMacroGenerator:
    def __init__(self, context: Optional[Dict[str, Any]] = None) -> None:
        self.context: Optional[Dict[str, Any]] = context

    def get_template(self) -> jinja2.Template:
        raise NotImplementedError("get_template not implemented!")

    def get_name(self) -> str:
        raise NotImplementedError("get_name not implemented!")

    def get_macro(self) -> Callable:
        name = self.get_name()
        template = self.get_template()
        # make the module. previously we set both vars and local, but that's
        # redundant: They both end up in the same place
        # make_module is in jinja2.environment. It returns a TemplateModule
        module = template.make_module(vars=self.context, shared=False)
        macro = module.__dict__[get_dbt_macro_name(name)]

        return macro

    @contextmanager
    def exception_handler(self) -> Iterator[None]:
        try:
            yield
        except (TypeError, jinja2.exceptions.TemplateRuntimeError) as e:
            raise CaughtMacroError(e)

    def call_macro(self, *args: Any, **kwargs: Any) -> Any:
        # called from __call__ methods
        if self.context is None:
            raise DbtInternalError("Context is still None in call_macro!")
        assert self.context is not None

        macro = self.get_macro()

        with self.exception_handler():
            try:
                return macro(*args, **kwargs)
            except MacroReturn as e:
                return e.value


class CallableMacroGenerator(BaseMacroGenerator):
    def __init__(
        self,
        macro: MacroProtocol,
        context: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(context)
        self.macro = macro

    def get_template(self) -> jinja2.Template:
        return template_cache.get_node_template(self.macro)

    def get_name(self) -> str:
        return self.macro.name

    @contextmanager
    def exception_handler(self) -> Iterator[None]:
        try:
            yield
        except (TypeError, jinja2.exceptions.TemplateRuntimeError) as e:
            raise CaughtMacroErrorWithNodeError(exc=e, node=self.macro)
        except CompilationError as e:
            e.stack.append(self.macro)
            raise e

    # this makes MacroGenerator objects callable like functions
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.call_macro(*args, **kwargs)


class MaterializationExtension(jinja2.ext.Extension):
    tags = ["materialization"]

    def parse(self, parser: jinja2.parser.Parser) -> _ParseReturn:
        node = jinja2.nodes.Macro(lineno=next(parser.stream).lineno)
        materialization_name = parser.parse_assign_target(name_only=True).name

        adapter_name = "default"
        node.args = []
        node.defaults = []

        while parser.stream.skip_if("comma"):
            target = parser.parse_assign_target(name_only=True)

            if target.name == "default":
                pass

            elif target.name == "adapter":
                parser.stream.expect("assign")
                value = parser.parse_expression()
                adapter_name = value.value

            elif target.name == "supported_languages":
                target.set_ctx("param")
                node.args.append(target)
                parser.stream.expect("assign")
                languages = parser.parse_expression()
                node.defaults.append(languages)

            else:
                raise MaterializationArgError(materialization_name, target.name)

        if SUPPORTED_LANG_ARG not in node.args:
            node.args.append(SUPPORTED_LANG_ARG)
            node.defaults.append(jinja2.nodes.List([jinja2.nodes.Const("sql")]))

        node.name = get_materialization_macro_name(materialization_name, adapter_name)

        node.body = parser.parse_statements(("name:endmaterialization",), drop_needle=True)

        return node


class DocumentationExtension(jinja2.ext.Extension):
    tags = ["docs"]

    def parse(self, parser: jinja2.parser.Parser) -> _ParseReturn:
        node = jinja2.nodes.Macro(lineno=next(parser.stream).lineno)
        docs_name = parser.parse_assign_target(name_only=True).name

        node.args = []
        node.defaults = []
        node.name = get_docs_macro_name(docs_name)
        node.body = parser.parse_statements(("name:enddocs",), drop_needle=True)
        return node


class TestExtension(jinja2.ext.Extension):
    tags = ["test"]

    def parse(self, parser: jinja2.parser.Parser) -> _ParseReturn:
        node = jinja2.nodes.Macro(lineno=next(parser.stream).lineno)
        test_name = parser.parse_assign_target(name_only=True).name

        parser.parse_signature(node)
        node.name = get_test_macro_name(test_name)
        node.body = parser.parse_statements(("name:endtest",), drop_needle=True)
        return node


def _is_dunder_name(name: str) -> bool:
    return name.startswith("__") and name.endswith("__")


def create_undefined(node: Optional[_NodeProtocol] = None) -> Type[jinja2.Undefined]:
    class Undefined(jinja2.Undefined):
        def __init__(
            self,
            hint: Optional[str] = None,
            obj: Any = None,
            name: Optional[str] = None,
            exc: Any = None,
        ) -> None:
            super().__init__(hint=hint, name=name)
            self.node = node
            self.name = name
            self.hint = hint
            # jinja uses these for safety, so we have to override them.
            # see https://github.com/pallets/jinja/blob/master/jinja2/sandbox.py#L332-L339 # noqa
            self.unsafe_callable = False
            self.alters_data = False

        def __getitem__(self, name: Any) -> "Undefined":
            # Propagate the undefined value if a caller accesses this as if it
            # were a dictionary
            return self

        def __getattr__(self, name: str) -> "Undefined":
            if name == "name" or _is_dunder_name(name):
                raise AttributeError(
                    "'{}' object has no attribute '{}'".format(type(self).__name__, name)
                )

            self.name = name

            return self.__class__(hint=self.hint, name=self.name)

        def __call__(self, *args: Any, **kwargs: Any) -> "Undefined":
            return self

        def __reduce__(self) -> NoReturn:
            raise UndefinedCompilationError(name=self.name or "unknown", node=node)

    return Undefined


def is_list(value):
    return isinstance(value, list)


NATIVE_FILTERS: Dict[str, Callable[[Any], Any]] = {
    "as_text": TextMarker,
    "as_bool": BoolMarker,
    "as_native": NativeMarker,
    "as_number": NumberMarker,
    "is_list": is_list,
}


TEXT_FILTERS: Dict[str, Callable[[Any], Any]] = {
    "as_text": lambda x: x,
    "as_bool": lambda x: x,
    "as_native": lambda x: x,
    "as_number": lambda x: x,
    "is_list": is_list,
}


def get_environment(
    node: Optional[_NodeProtocol] = None,
    capture_macros: bool = False,
    native: bool = False,
) -> jinja2.Environment:
    args: Dict[str, List[Union[str, Type[jinja2.ext.Extension]]]] = {
        "extensions": ["jinja2.ext.do", "jinja2.ext.loopcontrols"]
    }

    if capture_macros:
        args["undefined"] = create_undefined(node)  # type: ignore

    args["extensions"].append(MaterializationExtension)
    args["extensions"].append(DocumentationExtension)
    args["extensions"].append(TestExtension)

    env_cls: Type[jinja2.Environment]
    if native:
        env_cls = NativeSandboxEnvironment
        filters = NATIVE_FILTERS
    else:
        env_cls = MacroFuzzEnvironment
        filters = TEXT_FILTERS

    env = env_cls(**args)
    env.filters.update(filters)

    return env


@contextmanager
def catch_jinja(node: Optional[_NodeProtocol] = None) -> Iterator[None]:
    try:
        yield
    except jinja2.exceptions.TemplateSyntaxError as e:
        e.translated = False
        raise CompilationError(str(e), node) from e
    except jinja2.exceptions.UndefinedError as e:
        raise UndefinedMacroError(str(e), node) from e
    except CompilationError as exc:
        exc.add_node(node)
        raise
    except DbtRuntimeError:
        # Propagate dbt exception raised during jinja compilation
        raise
    except Exception as e:
        # Raise any non-dbt exceptions as CompilationError
        raise CompilationError(str(e), node) from e


_TESTING_PARSE_CACHE: Dict[str, jinja2.nodes.Template] = {}


def parse(string: Any) -> jinja2.nodes.Template:
    str_string = str(string)
    if test_caching_enabled() and str_string in _TESTING_PARSE_CACHE:
        return _TESTING_PARSE_CACHE[str_string]

    with catch_jinja():
        parsed: jinja2.nodes.Template = get_environment().parse(str(string))
        if test_caching_enabled():
            _TESTING_PARSE_CACHE[str_string] = parsed
        return parsed


def get_template(
    string: str,
    ctx: Dict[str, Any],
    node: Optional[_NodeProtocol] = None,
    capture_macros: bool = False,
    native: bool = False,
) -> jinja2.Template:
    with catch_jinja(node):
        env = get_environment(node, capture_macros, native=native)

        template_source = str(string)
        return env.from_string(template_source, globals=ctx)


def render_template(
    template: jinja2.Template, ctx: Dict[str, Any], node: Optional[_NodeProtocol] = None
) -> str:
    with catch_jinja(node):
        return template.render(ctx)


_TESTING_BLOCKS_CACHE: Dict[int, List[Union[BlockData, BlockTag]]] = {}


def _get_blocks_hash(text: str, allowed_blocks: Optional[Set[str]], collect_raw_data: bool) -> int:
    """Provides a hash function over the arguments to extract_toplevel_blocks, in order to support caching."""
    allowed_blocks = allowed_blocks or set()
    allowed_tuple = tuple(sorted(allowed_blocks) or [])
    return text.__hash__() + allowed_tuple.__hash__() + collect_raw_data.__hash__()


def extract_toplevel_blocks(
    text: str,
    allowed_blocks: Optional[Set[str]] = None,
    collect_raw_data: bool = True,
    warning_callback: Optional[Callable[[ExtractWarning], None]] = None,
) -> List[Union[BlockData, BlockTag]]:
    """Extract the top-level blocks with matching block types from a jinja file.

    Includes some special handling for block nesting.

    :param text: The data to extract blocks from.
    :param allowed_blocks: The names of the blocks to extract from the file.
        They may not be nested within if/for blocks. If None, use the default
        values.
    :param collect_raw_data: If set, raw data between matched blocks will also
        be part of the results, as `BlockData` objects. They have a
        `block_type_name` field of `'__dbt_data'` and will never have a
        `block_name`.
    :param warning_callback: An optional callback that will be called if there
        are recoverable issues detected in the template.
    :return: A list of `BlockTag`s matching the allowed block types and (if
        `collect_raw_data` is `True`) `BlockData` objects.
    """

    if test_caching_enabled():
        hash = _get_blocks_hash(text, allowed_blocks, collect_raw_data)
        if hash in _TESTING_BLOCKS_CACHE:
            return _TESTING_BLOCKS_CACHE[hash]

    tag_iterator = TagIterator(text)
    blocks = BlockIterator(tag_iterator, warning_callback).lex_for_blocks(
        allowed_blocks=allowed_blocks, collect_raw_data=collect_raw_data
    )

    if test_caching_enabled():
        hash = _get_blocks_hash(text, allowed_blocks, collect_raw_data)
        _TESTING_BLOCKS_CACHE[hash] = blocks

    return blocks



from _typeshed import Incomplete
from mashumaro.mixins.dict import DataClassDictMixin as DataClassDictMixin
from typing import Any, Callable, Dict, TypeVar, Union

T = TypeVar("T", bound="DataClassYAMLMixin")
EncodedData = Union[str, bytes]
Encoder = Callable[[Any], EncodedData]
Decoder = Callable[[EncodedData], Dict[Any, Any]]
DefaultLoader: Incomplete
DefaultDumper: Incomplete

def default_encoder(data: Any) -> EncodedData: ...
def default_decoder(data: EncodedData) -> Dict[Any, Any]: ...

class DataClassYAMLMixin(DataClassDictMixin):
    def to_yaml(self, encoder: Encoder = ..., **to_dict_kwargs: Any) -> EncodedData: ...
    @classmethod
    def from_yaml(
        cls, data: EncodedData, decoder: Decoder = ..., **from_dict_kwargs: Any
    ) -> T: ...



# necessary for annotating constructors
from __future__ import annotations

from dataclasses import dataclass, Field

from itertools import chain
from typing import Any, Callable, Dict, Iterator, List, Type, TypeVar

from dbt_common.contracts.config.metadata import Metadata
from dbt_common.exceptions import CompilationError, DbtInternalError
from dbt_common.contracts.config.properties import AdditionalPropertiesAllowed
from dbt_common.contracts.util import Replaceable

T = TypeVar("T", bound="BaseConfig")


@dataclass
class BaseConfig(AdditionalPropertiesAllowed, Replaceable):
    # enable syntax like: config['key']
    def __getitem__(self, key: str) -> Any:
        return self.get(key)

    # like doing 'get' on a dictionary
    def get(self, key: str, default: Any = None) -> Any:
        if hasattr(self, key):
            return getattr(self, key)
        elif key in self._extra:
            return self._extra[key]
        else:
            return default

    # enable syntax like: config['key'] = value
    def __setitem__(self, key: str, value) -> None:
        if hasattr(self, key):
            setattr(self, key, value)
        else:
            self._extra[key] = value

    def __delitem__(self, key: str) -> None:
        if hasattr(self, key):
            msg = (
                'Error, tried to delete config key "{}": Cannot delete ' "built-in keys"
            ).format(key)
            raise CompilationError(msg)
        else:
            del self._extra[key]

    def _content_iterator(self, include_condition: Callable[[Field[Any]], bool]) -> Iterator[str]:
        seen = set()
        for fld, _ in self._get_fields():
            seen.add(fld.name)
            if include_condition(fld):
                yield fld.name

        for key in self._extra:
            if key not in seen:
                seen.add(key)
                yield key

    def __iter__(self) -> Iterator[str]:
        yield from self._content_iterator(include_condition=lambda f: True)

    def __len__(self) -> int:
        return len(self._get_fields()) + len(self._extra)

    @staticmethod
    def compare_key(
        unrendered: Dict[str, Any],
        other: Dict[str, Any],
        key: str,
    ) -> bool:
        if key not in unrendered and key not in other:
            return True
        elif key not in unrendered and key in other:
            return False
        elif key in unrendered and key not in other:
            return False
        else:
            return bool(unrendered[key] == other[key])

    @classmethod
    def same_contents(cls, unrendered: Dict[str, Any], other: Dict[str, Any]) -> bool:
        """This is like __eq__, except it ignores some fields."""
        seen = set()
        for fld, target_name in cls._get_fields():
            key = target_name
            seen.add(key)
            if CompareBehavior.should_include(fld):
                if not cls.compare_key(unrendered, other, key):
                    return False

        for key in chain(unrendered, other):
            if key not in seen:
                seen.add(key)
                if not cls.compare_key(unrendered, other, key):
                    return False
        return True

    # This is used in 'merge_config_dicts' to create the combined orig_dict.
    # Note: "clobber" fields aren't defined, because that's the default.
    #    "access" is currently the only Clobber field.
    # This shouldn't really be defined  here. It would be better to have it
    # associated with the config definitions, but at the point we use it, we
    # don't know which config we're dealing with.
    mergebehavior = {
        "append": ["pre-hook", "pre_hook", "post-hook", "post_hook", "tags", "packages"],
        "update": [
            "quoting",
            "column_types",
            "meta",
            "docs",
            "contract",
        ],
        "dict_key_append": ["grants"],
        "object": ["snapshot_meta_column_names"],
    }

    @classmethod
    def _merge_dicts(cls, src: Dict[str, Any], data: Dict[str, Any]) -> Dict[str, Any]:
        """Mutate input to return merge results.

        Find all the items in data that match a target_field on this class,
        and merge them with the data found in `src` for target_field, using the
        field's specified merge behavior. Matching items will be removed from
        `data` (but _not_ `src`!).

        Returns a dict with the merge results.

        That means this method mutates its input! Any remaining values in data
        were not merged.
        """
        result = {}

        for fld, target_field in cls._get_fields():
            if target_field not in data:
                continue

            data_attr = data.pop(target_field)
            if target_field not in src:
                result[target_field] = data_attr
                continue

            merge_behavior = MergeBehavior.from_field(fld)
            self_attr = src[target_field]

            result[target_field] = _merge_field_value(
                merge_behavior=merge_behavior,
                self_value=self_attr,
                other_value=data_attr,
            )
        return result

    def update_from(
        self: T, data: Dict[str, Any], config_cls: Type[BaseConfig], validate: bool = True
    ) -> T:
        """Update and validate config given a dict.

        Given a dict of keys, update the current config from them, validate
        it, and return a new config with the updated values
        """
        dct = self.to_dict(omit_none=False)

        self_merged = self._merge_dicts(dct, data)
        dct.update(self_merged)

        adapter_merged = config_cls._merge_dicts(dct, data)
        dct.update(adapter_merged)

        # any remaining fields must be "clobber"
        dct.update(data)

        # any validation failures must have come from the update
        if validate:
            self.validate(dct)
        return self.from_dict(dct)

    def finalize_and_validate(self: T) -> T:
        dct = self.to_dict(omit_none=False)
        self.validate(dct)
        return self.from_dict(dct)


class MergeBehavior(Metadata):
    Append = 1
    Update = 2
    Clobber = 3
    DictKeyAppend = 4
    Object = 5

    @classmethod
    def default_field(cls) -> "MergeBehavior":
        return cls.Clobber

    @classmethod
    def metadata_key(cls) -> str:
        return "merge"


class CompareBehavior(Metadata):
    Include = 1
    Exclude = 2

    @classmethod
    def default_field(cls) -> "CompareBehavior":
        return cls.Include

    @classmethod
    def metadata_key(cls) -> str:
        return "compare"

    @classmethod
    def should_include(cls, fld: Field[Any]) -> bool:
        return cls.from_field(fld) == cls.Include


def _listify(value: Any) -> List[Any]:
    if isinstance(value, list):
        return value[:]
    else:
        return [value]


# There are two versions of this code. The one here is for config
# objects which can get the "MergeBehavior" from the field in the class,
# the one below in 'merge_config_dicts' (formerly in
# _add_config_call in core context_config.py) is for config_call dictionaries
# where we need to get the MergeBehavior from someplace else.
def _merge_field_value(
    merge_behavior: MergeBehavior,
    self_value: Any,
    other_value: Any,
) -> Any:
    if merge_behavior == MergeBehavior.Clobber:
        return other_value
    elif merge_behavior == MergeBehavior.Append:
        new_value = _listify(self_value) + _listify(other_value)
        return new_value
    elif merge_behavior == MergeBehavior.Update:
        if not isinstance(self_value, dict):
            raise DbtInternalError(f"expected dict, got {self_value}")
        if not isinstance(other_value, dict):
            raise DbtInternalError(f"expected dict, got {other_value}")
        value = self_value.copy()
        value.update(other_value)
        return value
    elif merge_behavior == MergeBehavior.DictKeyAppend:
        if not isinstance(self_value, dict):
            raise DbtInternalError(f"expected dict, got {self_value}")
        if not isinstance(other_value, dict):
            raise DbtInternalError(f"expected dict, got {other_value}")
        new_dict = {}
        for key in self_value.keys():
            new_dict[key] = _listify(self_value[key])
        for key in other_value.keys():
            extend = False
            new_key = key
            # This might start with a +, to indicate we should extend the list
            # instead of just clobbering it
            if new_key.startswith("+"):
                new_key = key.lstrip("+")
                extend = True
            if new_key in new_dict and extend:
                # extend the list
                value = other_value[key]
                new_dict[new_key].extend(_listify(value))
            else:
                # clobber the list
                new_dict[new_key] = _listify(other_value[key])
        return new_dict
    elif merge_behavior == MergeBehavior.Object:
        # All fields in classes with MergeBehavior.Object should have a default of None
        if not type(self_value).__name__ == type(other_value).__name__:
            raise DbtInternalError(
                f"got conflicting types: {type(self_value).__name__} and {type(other_value).__name__}"
            )
        new_value = self_value.copy()
        new_value.update(other_value)
        return new_value
    else:
        raise DbtInternalError(f"Got an invalid merge_behavior: {merge_behavior}")


# This is used in ContextConfig._add_config_call. It updates the orig_dict in place.
def merge_config_dicts(orig_dict: Dict[str, Any], new_dict: Dict[str, Any]) -> None:
    # orig_dict is already encountered configs, new_dict is new
    # This mirrors code in _merge_field_value in model_config.py which is similar but
    # operates on config objects.
    if orig_dict == {}:
        orig_dict.update(new_dict)
        return
    for k, v in new_dict.items():
        # MergeBehavior for post-hook and pre-hook is to collect all
        # values, instead of overwriting
        if k in BaseConfig.mergebehavior["append"]:
            if k in orig_dict:  # should always be a list here
                orig_dict[k] = _listify(orig_dict[k]) + _listify(v)
            else:
                orig_dict[k] = _listify(v)
        elif k in BaseConfig.mergebehavior["update"]:
            if not isinstance(v, dict):
                raise DbtInternalError(f"expected dict, got {v}")
            if k in orig_dict and isinstance(orig_dict[k], dict):
                orig_dict[k].update(v)
            else:
                orig_dict[k] = v
        elif k in BaseConfig.mergebehavior["dict_key_append"]:
            if not isinstance(v, dict):
                raise DbtInternalError(f"expected dict, got {v}")
            if k in orig_dict:  # should always be a dict
                for key in orig_dict[k].keys():
                    orig_dict[k][key] = _listify(orig_dict[k][key])
                for key, value in v.items():
                    extend = False
                    # This might start with a +, to indicate we should extend the list
                    # instead of just clobbering it. We don't want to remove the + here
                    # (like in the other method) because we want it preserved
                    if key.startswith("+"):
                        extend = True
                    if key in orig_dict[k] and extend:
                        # extend the list
                        orig_dict[k][key].extend(_listify(value))
                    else:
                        # clobber the list
                        orig_dict[k][key] = _listify(value)
            else:
                # This is always a dictionary
                orig_dict[k] = v
                # listify everything
                for key, value in orig_dict[k].items():
                    orig_dict[k][key] = _listify(value)
        elif k in BaseConfig.mergebehavior["object"]:
            if not isinstance(v, dict):
                raise DbtInternalError(f"expected dict, got {v}")
            if k not in orig_dict:
                orig_dict[k] = {}
            for obj_k, obj_v in v.items():
                orig_dict[k][obj_k] = obj_v
        else:  # Clobber
            orig_dict[k] = v



import inspect
from typing import Any, Dict, List, TypedDict

try:
    from typing import NotRequired
except ImportError:
    # NotRequired was introduced in Python 3.11
    # This is the suggested way to implement a TypedDict with optional arguments
    from typing import Optional as NotRequired

from dbt_common.events.functions import fire_event
from dbt_common.events.types import BehaviorChangeEvent
from dbt_common.exceptions import CompilationError, DbtInternalError


class BehaviorFlag(TypedDict):
    """
    Configuration used to create a BehaviorFlagRendered instance

    Args:
        name: the name of the behavior flag
        default: default setting, starts as False, becomes True after a bake-in period
        description: an additional message to send when the flag evaluates to False
        docs_url: the url to the relevant docs on docs.getdbt.com

    *Note*:
        While `description` and `docs_url` are both listed as `NotRequired`, at least one of them is required.
        This is validated when the flag is rendered in `BehaviorFlagRendered` below.
        The goal of this restriction is to provide the end user with context so they can make an informed decision
        about if, and when, to enable the behavior flag.
    """

    name: str
    default: bool
    source: NotRequired[str]
    description: NotRequired[str]
    docs_url: NotRequired[str]


class BehaviorFlagRendered:
    """
    A rendered behavior flag that gets used throughout dbt packages

    Args:
        flag: the configuration for the behavior flag
        user_overrides: a set of user settings, one of which may be an override on this behavior flag
    """

    fired: bool = False

    def __init__(self, flag: BehaviorFlag, user_overrides: Dict[str, Any]) -> None:
        self._validate(flag)

        self.name = flag["name"]
        self.setting = user_overrides.get(flag["name"], flag["default"])

        default_description = (
            f"""The behavior controlled by `{flag["name"]}` is currently turned off.\n"""
        )
        default_docs_url = "https://docs.getdbt.com/reference/global-configs/behavior-changes"
        self._behavior_change_event = BehaviorChangeEvent(
            flag_name=flag["name"],
            flag_source=flag.get("source", self._default_source()),
            description=flag.get("description", default_description),
            docs_url=flag.get("docs_url", default_docs_url),
        )

    @staticmethod
    def _validate(flag: BehaviorFlag) -> None:
        if flag.get("description") is None and flag.get("docs_url") is None:
            raise DbtInternalError(
                "Behavior change flags require at least one of `description` and `docs_url`."
            )

    @property
    def setting(self) -> bool:
        if self._setting is False and not self.fired:
            fire_event(self._behavior_change_event)
            self.fired = True
        return self._setting

    @setting.setter
    def setting(self, value: bool) -> None:
        self._setting = value

    @property
    def no_warn(self) -> bool:
        return self._setting

    @staticmethod
    def _default_source() -> str:
        """
        If the maintainer did not provide a source, default to the module that called this class.
        For adapters, this will likely be `dbt.adapters.<foo>.impl` for `dbt-foo`.
        """
        for frame in inspect.stack():
            if module := inspect.getmodule(frame[0]):
                if module.__name__ != __name__:
                    return module.__name__
        return "Unknown"

    def __bool__(self) -> bool:
        return self.setting


class Behavior:
    """
    A collection of behavior flags

    This is effectively a dictionary that supports dot notation for easy reference, e.g.:
        ```python
        if adapter.behavior.my_flag:
            ...

        if adapter.behavior.my_flag.no_warn:  # this will not fire the behavior change event
            ...
        ```
        ```jinja
        {% if adapter.behavior.my_flag %}
            ...
        {% endif %}

        {% if adapter.behavior.my_flag.no_warn %}  {# this will not fire the behavior change event #}
            ...
        {% endif %}
        ```

    Args:
        flags: a list of configurations, one for each behavior flag
        user_overrides: a set of user settings, which may include overrides on one or more of the behavior flags
    """

    _flags: List[BehaviorFlagRendered]

    def __init__(self, flags: List[BehaviorFlag], user_overrides: Dict[str, Any]) -> None:
        self._flags = [BehaviorFlagRendered(flag, user_overrides) for flag in flags]

    def __getattr__(self, name: str) -> BehaviorFlagRendered:
        for flag in self._flags:
            if flag.name == name:
                return flag
        raise CompilationError(f"The flag {name} has not been registered.")



from mashumaro.types import SerializationStrategy
from typing import Any, Callable, Dict, Optional, TypeVar, Union
from typing_extensions import Literal

NamedTupleDeserializationEngine = Literal["as_dict", "as_list"]
DateTimeDeserializationEngine = Literal["ciso8601", "pendulum"]
AnyDeserializationEngine = Literal[NamedTupleDeserializationEngine, DateTimeDeserializationEngine]

NamedTupleSerializationEngine = Literal["as_dict", "as_list"]
AnySerializationEngine = Union[NamedTupleSerializationEngine, OmitSerializationEngine]
OmitSerializationEngine = Literal["omit"]

T = TypeVar("T")

def field_options(
    serialize: Optional[Union[AnySerializationEngine, Callable[[Any], Any]]] = ...,
    deserialize: Optional[Union[AnyDeserializationEngine, Callable[[Any], Any]]] = ...,
    serialization_strategy: Optional[SerializationStrategy] = ...,
    alias: Optional[str] = ...,
) -> Dict[str, Any]: ...

class _PassThrough(SerializationStrategy):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
    def serialize(self, value: T) -> T: ...
    def deserialize(self, value: T) -> T: ...

pass_through: Any



from mashumaro.jsonschema.models import JSONSchema, Number
from typing import Dict, Set

class Annotation: ...
class Constraint(Annotation): ...
class NumberConstraint(Constraint): ...

class Minimum(NumberConstraint):
    value: Number
    def __init__(self, value) -> None: ...

class Maximum(NumberConstraint):
    value: Number
    def __init__(self, value) -> None: ...

class ExclusiveMinimum(NumberConstraint):
    value: Number
    def __init__(self, value) -> None: ...

class ExclusiveMaximum(NumberConstraint):
    value: Number
    def __init__(self, value) -> None: ...

class MultipleOf(NumberConstraint):
    value: Number
    def __init__(self, value) -> None: ...

class StringConstraint(Constraint): ...

class MinLength(StringConstraint):
    value: int
    def __init__(self, value) -> None: ...

class MaxLength(StringConstraint):
    value: int
    def __init__(self, value) -> None: ...

class Pattern(StringConstraint):
    value: str
    def __init__(self, value) -> None: ...

class ArrayConstraint(Constraint): ...

class MinItems(ArrayConstraint):
    value: int
    def __init__(self, value) -> None: ...

class MaxItems(ArrayConstraint):
    value: int
    def __init__(self, value) -> None: ...

class UniqueItems(ArrayConstraint):
    value: bool
    def __init__(self, value) -> None: ...

class Contains(ArrayConstraint):
    value: JSONSchema
    def __init__(self, value) -> None: ...

class MinContains(ArrayConstraint):
    value: int
    def __init__(self, value) -> None: ...

class MaxContains(ArrayConstraint):
    value: int
    def __init__(self, value) -> None: ...

class ObjectConstraint(Constraint): ...

class MaxProperties(ObjectConstraint):
    value: int
    def __init__(self, value) -> None: ...

class MinProperties(ObjectConstraint):
    value: int
    def __init__(self, value) -> None: ...

class DependentRequired(ObjectConstraint):
    value: Dict[str, Set[str]]
    def __init__(self, value) -> None: ...



from mashumaro.core.const import Sentinel
from mashumaro.types import SerializationStrategy
from typing import Callable, Dict, Union
from typing_extensions import Literal

SerializationStrategyValueType = Union[SerializationStrategy, Dict[str, Union[str, Callable]]]

class Dialect:
    serialization_strategy: Dict[str, SerializationStrategyValueType]
    omit_none: Union[bool, Sentinel.MISSING]



from _typeshed import Incomplete
from functools import cached_property
from mashumaro.core.const import PEP_585_COMPATIBLE as PEP_585_COMPATIBLE
from mashumaro.core.meta.code.builder import CodeBuilder as CodeBuilder
from mashumaro.core.meta.helpers import (
    get_type_origin as get_type_origin,
    is_annotated as is_annotated,
    is_generic as is_generic,
    type_name as type_name,
)
from mashumaro.exceptions import (
    UnserializableDataError as UnserializableDataError,
    UnserializableField as UnserializableField,
)
from typing import Any, Dict, Mapping, Optional, Sequence, Type, TypeVar
from typing_extensions import TypeAlias

cached_property = property
NoneType: Incomplete
Expression: TypeAlias
P: Incomplete
T = TypeVar("T")

class ExpressionWrapper:
    expression: Incomplete
    def __init__(self, expression: str) -> None: ...

PROPER_COLLECTION_TYPES: Dict[Type, str]

class FieldContext:
    name: str
    metadata: Mapping
    def copy(self, **changes: Any) -> FieldContext: ...
    def __init__(self, name, metadata) -> None: ...

class ValueSpec:
    type: Type
    origin_type: Type
    expression: Expression
    builder: CodeBuilder
    field_ctx: FieldContext
    could_be_none: bool
    annotated_type: Optional[Type]
    def __setattr__(self, key: str, value: Any) -> None: ...
    def copy(self, **changes: Any) -> ValueSpec: ...
    @cached_property
    def annotations(self) -> Sequence[str]: ...
    def __init__(
        self, type, expression, builder, field_ctx, could_be_none, annotated_type
    ) -> None: ...

ValueSpecExprCreator: TypeAlias

class Registry:
    def register(self, function: ValueSpecExprCreator) -> ValueSpecExprCreator: ...
    def get(self, spec: ValueSpec) -> Expression: ...
    def __init__(self, _registry) -> None: ...

def ensure_generic_collection(spec: ValueSpec) -> bool: ...
def ensure_collection_type_args_supported(
    collection_type: Type, type_args: Sequence[Type]
) -> bool: ...
def ensure_generic_collection_subclass(spec: ValueSpec, *checked_types: Type) -> bool: ...
def ensure_generic_mapping(spec: ValueSpec, args: Sequence[Type], checked_type: Type) -> bool: ...
def expr_or_maybe_none(spec: ValueSpec, new_expr: Expression) -> Expression: ...
def random_hex() -> str: ...
def clean_id(value: str) -> str: ...



from dataclasses import Field
from enum import Enum
from typing import TypeVar, Type, Optional, Dict, Any

from dbt_common.exceptions import DbtInternalError

M = TypeVar("M", bound="Metadata")


class Metadata(Enum):
    @classmethod
    def from_field(cls: Type[M], fld: Field) -> M:
        default = cls.default_field()
        key = cls.metadata_key()

        return _get_meta_value(cls, fld, key, default)

    def meta(self, existing: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        key = self.metadata_key()
        return _set_meta_value(self, key, existing)

    @classmethod
    def default_field(cls) -> "Metadata":
        raise NotImplementedError("Not implemented")

    @classmethod
    def metadata_key(cls) -> str:
        raise NotImplementedError("Not implemented")


def _get_meta_value(cls: Type[M], fld: Field, key: str, default: Any) -> M:
    # a metadata field might exist. If it does, it might have a matching key.
    # If it has both, make sure the value is valid and return it. If it
    # doesn't, return the default.
    if fld.metadata:
        value = fld.metadata.get(key, default)
    else:
        value = default

    try:
        return cls(value)
    except ValueError as exc:
        raise DbtInternalError(f"Invalid {cls} value: {value}") from exc


def _set_meta_value(obj: M, key: str, existing: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    if existing is None:
        result = {}
    else:
        result = existing.copy()
    result.update({key: obj})
    return result


class ShowBehavior(Metadata):
    Show = 1
    Hide = 2

    @classmethod
    def default_field(cls) -> "ShowBehavior":
        return cls.Show

    @classmethod
    def metadata_key(cls) -> str:
        return "show_hide"

    @classmethod
    def should_show(cls, fld: Field) -> bool:
        return cls.from_field(fld) == cls.Show



from dataclasses import dataclass
from typing import Dict, Optional, Union, NamedTuple

from dbt_common.dataclass_schema import dbtClassMixin
from dbt_common.utils.formatting import lowercase


@dataclass
class StatsItem(dbtClassMixin):
    id: str
    label: str
    value: Union[bool, str, float, None]
    include: bool
    description: Optional[str] = None


StatsDict = Dict[str, StatsItem]


@dataclass
class TableMetadata(dbtClassMixin):
    type: str
    schema: str
    name: str
    database: Optional[str] = None
    comment: Optional[str] = None
    owner: Optional[str] = None


CatalogKey = NamedTuple(
    "CatalogKey", [("database", Optional[str]), ("schema", str), ("name", str)]
)


@dataclass
class ColumnMetadata(dbtClassMixin):
    type: str
    index: int
    name: str
    comment: Optional[str] = None


ColumnMap = Dict[str, ColumnMetadata]


@dataclass
class CatalogTable(dbtClassMixin):
    metadata: TableMetadata
    columns: ColumnMap
    stats: StatsDict
    # the same table with two unique IDs will just be listed two times
    unique_id: Optional[str] = None

    def key(self) -> CatalogKey:
        return CatalogKey(
            lowercase(self.metadata.database),
            self.metadata.schema.lower(),
            self.metadata.name.lower(),
        )



import enum
from _typeshed import Incomplete

PY_37: Incomplete
PY_38: Incomplete
PY_39: Incomplete
PY_310: Incomplete
PY_311_MIN: Incomplete
PY_310_MIN: Incomplete
PY_39_MIN: Incomplete
PY_38_MIN: Incomplete
PY_37_MIN: Incomplete
PEP_585_COMPATIBLE = PY_39_MIN
PEP_586_COMPATIBLE = PY_38_MIN

class Sentinel(enum.Enum):
    MISSING: Incomplete



from _typeshed import Incomplete
from mashumaro.core.meta.helpers import type_name as type_name
from typing import Any, Optional, Type

class MissingField(LookupError):
    field_name: Incomplete
    field_type: Incomplete
    holder_class: Incomplete
    def __init__(self, field_name: str, field_type: Type, holder_class: Type) -> None: ...
    @property
    def field_type_name(self) -> str: ...
    @property
    def holder_class_name(self) -> str: ...

class UnserializableDataError(TypeError): ...

class UnserializableField(UnserializableDataError):
    field_name: Incomplete
    field_type: Incomplete
    holder_class: Incomplete
    msg: Incomplete
    def __init__(
        self, field_name: str, field_type: Type, holder_class: Type, msg: Optional[str] = ...
    ) -> None: ...
    @property
    def field_type_name(self) -> str: ...
    @property
    def holder_class_name(self) -> str: ...

class UnsupportedSerializationEngine(UnserializableField):
    def __init__(
        self, field_name: str, field_type: Type, holder_class: Type, engine: Any
    ) -> None: ...

class UnsupportedDeserializationEngine(UnserializableField):
    def __init__(
        self, field_name: str, field_type: Type, holder_class: Type, engine: Any
    ) -> None: ...

class InvalidFieldValue(ValueError):
    field_name: Incomplete
    field_type: Incomplete
    field_value: Incomplete
    holder_class: Incomplete
    msg: Incomplete
    def __init__(
        self,
        field_name: str,
        field_type: Type,
        field_value: Any,
        holder_class: Type,
        msg: Optional[str] = ...,
    ) -> None: ...
    @property
    def field_type_name(self) -> str: ...
    @property
    def holder_class_name(self) -> str: ...

class MissingDiscriminatorError(LookupError):
    field_name: Incomplete
    def __init__(self, field_name: str) -> None: ...

class SuitableVariantNotFoundError(ValueError):
    variants_type: Incomplete
    discriminator_name: Incomplete
    discriminator_value: Incomplete
    def __init__(
        self,
        variants_type: Type,
        discriminator_name: Optional[str] = ...,
        discriminator_value: Any = ...,
    ) -> None: ...

class BadHookSignature(TypeError): ...

class ThirdPartyModuleNotFoundError(ModuleNotFoundError):
    module_name: Incomplete
    field_name: Incomplete
    holder_class: Incomplete
    def __init__(self, module_name: str, field_name: str, holder_class: Type) -> None: ...
    @property
    def holder_class_name(self) -> str: ...

class UnresolvedTypeReferenceError(NameError):
    holder_class: Incomplete
    name: Incomplete
    def __init__(self, holder_class: Type, unresolved_type_name: str) -> None: ...
    @property
    def holder_class_name(self) -> str: ...

class BadDialect(ValueError): ...



from mashumaro.config import BaseConfig
from mashumaro.jsonschema.annotations import Annotation
from mashumaro.jsonschema.models import Context, JSONSchema
from typing import Any, Callable, Iterable, List, Mapping, Optional, Tuple, Type, Union

class Instance:
    type: Type
    name: Optional[str]
    origin_type: Type
    annotations: List[Annotation]
    @property
    def metadata(self) -> Mapping[str, Any]: ...
    @property
    def alias(self) -> Optional[str]: ...
    @property
    def holder_class(self) -> Optional[Type]: ...
    def copy(self, **changes: Any) -> Instance: ...
    def __post_init__(self) -> None: ...
    def update_type(self, new_type: Type) -> None: ...
    def fields(self) -> Iterable[Tuple[str, Type, bool, Any]]: ...
    def get_overridden_serialization_method(self) -> Optional[Union[Callable, str]]: ...
    def get_config(self) -> Type[BaseConfig]: ...
    def __init__(self, type, name, __builder) -> None: ...

class InstanceSchemaCreatorRegistry:
    def register(self, func: InstanceSchemaCreator) -> InstanceSchemaCreator: ...
    def iter(self) -> Iterable[InstanceSchemaCreator]: ...
    def __init__(self, _registry) -> None: ...

class EmptyJSONSchema(JSONSchema):
    def __init__(
        self,
        schema,
        type,
        enum,
        const,
        format,
        title,
        description,
        anyOf,
        reference,
        definitions,
        default,
        deprecated,
        examples,
        properties,
        patternProperties,
        additionalProperties,
        propertyNames,
        prefixItems,
        items,
        contains,
        multipleOf,
        maximum,
        exclusiveMaximum,
        minimum,
        exclusiveMinimum,
        maxLength,
        minLength,
        pattern,
        maxItems,
        minItems,
        uniqueItems,
        maxContains,
        minContains,
        maxProperties,
        minProperties,
        required,
        dependentRequired,
    ) -> None: ...

def get_schema(instance: Instance, ctx: Context, with_dialect_uri: bool = ...) -> JSONSchema: ...



from dbt_common.utils.encoding import md5, JSONEncoder, ForgivingJSONEncoder

from dbt_common.utils.casting import (
    cast_to_str,
    cast_to_int,
    cast_dict_to_dict_of_strings,
)

from dbt_common.utils.dict import (
    AttrDict,
    filter_null_values,
    merge,
    deep_merge,
    deep_merge_item,
    deep_map_render,
)

from dbt_common.utils.executor import executor

from dbt_common.utils.jinja import (
    get_dbt_macro_name,
    get_docs_macro_name,
    get_materialization_macro_name,
    get_test_macro_name,
    MACRO_PREFIX,
)



from dbt_common.events.base_types import (
    DebugLevel,
    InfoLevel,
    WarnLevel,
)
from dbt_common.ui import warning_tag


# The classes in this file represent the data necessary to describe a
# particular event to both human readable logs, and machine reliable
# event streams. classes extend superclasses that indicate what
# destinations they are intended for, which mypy uses to enforce
# that the necessary methods are defined.


# Event codes have prefixes which follow this table
#
# | Code |     Description     |
# |:----:|:-------------------:|
# | A    | Pre-project loading |
# | D    | Deprecations        |
# | E    | DB adapter          |
# | I    | Project parsing     |
# | M    | Deps generation     |
# | P    | Artifacts           |
# | Q    | Node execution      |
# | W    | Node testing        |
# | Z    | Misc                |
# | T    | Test only           |
#
# The basic idea is that event codes roughly translate to the natural order of running a dbt task


# =======================================================
# D - Deprecations
# =======================================================


class BehaviorChangeEvent(WarnLevel):
    def code(self) -> str:
        return "D018"

    def message(self) -> str:
        return warning_tag(
            f"{self.description}\n"
            f"You may opt into the new behavior sooner by setting `flags.{self.flag_name}` to `True` in `dbt_project.yml`.\n"
            f"Visit {self.docs_url} for more information."
        )


# =======================================================
# M - Deps generation
# =======================================================


class RetryExternalCall(DebugLevel):
    def code(self) -> str:
        return "M020"

    def message(self) -> str:
        return f"Retrying external call. Attempt: {self.attempt} Max attempts: {self.max}"


class RecordRetryException(DebugLevel):
    def code(self) -> str:
        return "M021"

    def message(self) -> str:
        return f"External call exception: {self.exc}"


# =======================================================
# Z - Misc
# =======================================================


class SystemCouldNotWrite(DebugLevel):
    def code(self) -> str:
        return "Z005"

    def message(self) -> str:
        return (
            f"Could not write to path {self.path}({len(self.path)} characters): "
            f"{self.reason}\nexception: {self.exc}"
        )


class SystemExecutingCmd(DebugLevel):
    def code(self) -> str:
        return "Z006"

    def message(self) -> str:
        return f'Executing "{" ".join(self.cmd)}"'


class SystemStdOut(DebugLevel):
    def code(self) -> str:
        return "Z007"

    def message(self) -> str:
        return f'STDOUT: "{str(self.bmsg)}"'


class SystemStdErr(DebugLevel):
    def code(self) -> str:
        return "Z008"

    def message(self) -> str:
        return f'STDERR: "{str(self.bmsg)}"'


class SystemReportReturnCode(DebugLevel):
    def code(self) -> str:
        return "Z009"

    def message(self) -> str:
        return f"command return code={self.returncode}"


# We use events to create console output, but also think of them as a sequence of important and
# meaningful occurrences to be used for debugging and monitoring. The Formatting event eases
# the tension between these two goals by allowing empty lines, heading separators, and other
# formatting to be written to the console, while they can be ignored for other purposes. For
# general information that isn't simple formatting, the Note event should be used instead.


class Formatting(InfoLevel):
    def code(self) -> str:
        return "Z017"

    def message(self) -> str:
        return self.msg


class Note(InfoLevel):
    """Unstructured events.

    The Note event provides a way to log messages which aren't likely to be
    useful as more structured events. For console formatting text like empty
    lines and separator bars, use the Formatting event instead.
    """

    def code(self) -> str:
        return "Z050"

    def message(self) -> str:
        return self.msg


class PrintEvent(InfoLevel):
    # Use this event to skip any formatting and just print a message
    # This event will get to stdout even if the logger is set to ERROR
    # This is to support commands that want --quiet option but also log something to stdout
    def code(self) -> str:
        return "Z052"

    def message(self) -> str:
        return self.msg


class RecordReplayIssue(InfoLevel):
    """General event for reporting record/replay issues at runtime."""

    def code(self) -> str:
        return "Z053"

    def message(self) -> str:
        return self.msg



from _typeshed import Incomplete

class JSONSchemaDialect:
    uri: str
    definitions_root_pointer: str
    all_refs: bool
    def __init__(self, uri, definitions_root_pointer, all_refs) -> None: ...

class JSONSchemaDraft202012Dialect(JSONSchemaDialect):
    uri: str
    definitions_root_pointer: str
    all_refs: bool
    def __init__(self, uri, definitions_root_pointer, all_refs) -> None: ...

class OpenAPISchema31Dialect(JSONSchemaDialect):
    uri: str
    definitions_root_pointer: str
    all_refs: bool
    def __init__(self, uri, definitions_root_pointer, all_refs) -> None: ...

DRAFT_2020_12: Incomplete
OPEN_API_3_1: Incomplete



from mashumaro.mixins.dict import DataClassDictMixin as DataClassDictMixin
from typing import Any, Callable, Dict, TypeVar, Union, Type

EncodedData = Union[str, bytes, bytearray]
T = TypeVar("T", bound="DataClassJSONMixin")

class Encoder:
    def __call__(self, obj: Any, **kwargs: Any) -> EncodedData: ...

class Decoder:
    def __call__(self, s: EncodedData, **kwargs: Any) -> Dict[Any, Any]: ...

class DataClassJSONMixin(DataClassDictMixin):
    def to_json(self, encoder: Encoder = ..., **to_dict_kwargs: Any) -> EncodedData: ...
    @classmethod
    def from_json(
        cls: Type[T], data: EncodedData, decoder: Decoder = ..., **from_dict_kwargs: Any
    ) -> T: ...



from _typeshed import Incomplete
from mashumaro.jsonschema.dialects import JSONSchemaDialect
from mashumaro.jsonschema.models import Context, JSONSchema
from mashumaro.mixins.json import DataClassJSONMixin
from typing import Any, Dict, List, Optional, Type

def build_json_schema(
    instance_type: Type,
    context: Optional[Context] = ...,
    with_definitions: bool = ...,
    all_refs: Optional[bool] = ...,
    with_dialect_uri: bool = ...,
    dialect: Optional[JSONSchemaDialect] = ...,
    ref_prefix: Optional[str] = ...,
) -> JSONSchema: ...

class JSONSchemaDefinitions(DataClassJSONMixin):
    definitions: Dict[str, JSONSchema]
    def __post_serialize__(self, d: Dict[Any, Any]) -> List[Dict[str, Any]]: ...  # type: ignore
    def __init__(self, definitions) -> None: ...

class JSONSchemaBuilder:
    context: Incomplete
    def __init__(
        self,
        dialect: JSONSchemaDialect = ...,
        all_refs: Optional[bool] = ...,
        ref_prefix: Optional[str] = ...,
    ) -> None: ...
    def build(self, instance_type: Type) -> JSONSchema: ...
    def get_definitions(self) -> JSONSchemaDefinitions: ...



_TEST_CACHING_ENABLED: bool = False


def test_caching_enabled() -> bool:
    return _TEST_CACHING_ENABLED


def enable_test_caching() -> None:
    global _TEST_CACHING_ENABLED
    _TEST_CACHING_ENABLED = True


def disable_test_caching() -> None:
    global _TEST_CACHING_ENABLED
    _TEST_CACHING_ENABLED = False



from dataclasses import dataclass, field
from typing import Dict, Any, Optional

from dbt_common.dataclass_schema import ExtensibleDbtClassMixin, dbtClassMixin


class AdditionalPropertiesMixin(dbtClassMixin):
    """Make this class an extensible property.

    The underlying class definition must include a type definition for a field
    named '_extra' that is of type `Dict[str, Any]`.
    """

    ADDITIONAL_PROPERTIES = True

    # This takes attributes in the dictionary that are
    # not in the class definitions and puts them in an
    # _extra dict in the class
    @classmethod
    def __pre_deserialize__(cls, data):
        # dir() did not work because fields with
        # metadata settings are not found
        # The original version of this would create the
        # object first and then update extra with the
        # extra keys, but that won't work here, so
        # we're copying the dict so we don't insert the
        # _extra in the original data. This also requires
        # that Mashumaro actually build the '_extra' field
        cls_keys = cls._get_field_names()
        new_dict = {}
        for key, value in data.items():
            # The pre-hook/post-hook mess hasn't been converted yet... That happens in
            # the super().__pre_deserialize__ below...
            if key not in cls_keys and key not in ["_extra", "pre-hook", "post-hook"]:
                if "_extra" not in new_dict:
                    new_dict["_extra"] = {}
                new_dict["_extra"][key] = value
            else:
                new_dict[key] = value
        data = new_dict
        data = super().__pre_deserialize__(data)
        return data

    def __post_serialize__(self, dct: Dict, context: Optional[Dict] = None):
        data = super().__post_serialize__(dct, context)
        data.update(self.extra)
        if "_extra" in data:
            del data["_extra"]
        return data

    def replace(self, **kwargs):
        dct = self.to_dict(omit_none=False)
        dct.update(kwargs)
        return self.from_dict(dct)

    @property
    def extra(self):
        return self._extra


@dataclass
class AdditionalPropertiesAllowed(AdditionalPropertiesMixin, ExtensibleDbtClassMixin):
    _extra: Dict[str, Any] = field(default_factory=dict)



from typing import Generator, Optional

class CodeLines:
    def __init__(self) -> None: ...
    def append(self, line: str) -> None: ...
    def indent(self, expr: Optional[str] = ...) -> Generator[None, None, None]: ...
    def as_text(self) -> str: ...
    def reset(self) -> None: ...



# never name this package "types", or mypy will crash in ugly ways

# necessary for annotating constructors
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Tuple, AbstractSet, Union
from typing import Callable, cast, Generic, Optional, TypeVar, List, NewType, Set

from dbt_common.dataclass_schema import (
    dbtClassMixin,
    ValidationError,
    StrEnum,
)
from dbt_common.events.base_types import BaseEvent

Port = NewType("Port", int)


class NVEnum(StrEnum):
    novalue = "novalue"

    def __eq__(self, other) -> bool:
        return isinstance(other, NVEnum)


@dataclass
class NoValue(dbtClassMixin):
    """Sometimes, you want a way to say none that isn't None!"""

    novalue: NVEnum = field(default_factory=lambda: NVEnum.novalue)


@dataclass
class IncludeExclude(dbtClassMixin):
    INCLUDE_ALL = ("all", "*")

    include: Union[str, List[str]]
    exclude: List[str] = field(default_factory=list)

    def __post_init__(self):
        if isinstance(self.include, str) and self.include not in self.INCLUDE_ALL:
            raise ValidationError(
                f"include must be one of {self.INCLUDE_ALL} or a list of strings"
            )

        if self.exclude and self.include not in self.INCLUDE_ALL:
            raise ValidationError(
                f"exclude can only be specified if include is one of {self.INCLUDE_ALL}"
            )

        if isinstance(self.include, list):
            self._validate_items(self.include)

        if isinstance(self.exclude, list):
            self._validate_items(self.exclude)

    def includes(self, item_name: str) -> bool:
        return (
            item_name in self.include or self.include in self.INCLUDE_ALL
        ) and item_name not in self.exclude

    def _validate_items(self, items: List[str]) -> None:
        pass


class WarnErrorOptions(IncludeExclude):
    """Deprecated, use WarnErrorOptionsV2 instead."""

    DEPRECATIONS = "Deprecations"

    def __init__(
        self,
        include: Union[str, List[str]],
        exclude: Optional[List[str]] = None,
        valid_error_names: Optional[Set[str]] = None,
        silence: Optional[List[str]] = None,
    ):
        self.silence = silence or []
        self._valid_error_names: Set[str] = valid_error_names or set()
        self._valid_error_names.add(self.DEPRECATIONS)
        super().__init__(include=include, exclude=(exclude or []))

        self._warn_error_options_v2 = WarnErrorOptionsV2(
            error=self.include,
            warn=self.exclude,
            silence=self.silence,
            valid_error_names=self._valid_error_names,
        )

    def __post_init__(self):
        # We don't want IncludeExclude's post_init to run, so we override it.
        # We are fine with just having the WarnErrorOptionsV2's post_init run on instantiation.
        pass

    def includes(self, item_name: Union[str, BaseEvent]) -> bool:
        return self._warn_error_options_v2.includes(item_name)

    def errors(self, item_name: Union[str, BaseEvent]) -> bool:
        """Exists for forward compatibility with WarnErrorOptionsV2."""
        return self._warn_error_options_v2.errors(item_name)

    def silenced(self, item_name: Union[str, BaseEvent]) -> bool:
        return self._warn_error_options_v2.silenced(item_name)


@dataclass
class WarnErrorOptionsV2(dbtClassMixin):
    """
    This class is used to configure the behavior of the warn_error feature (now part of fire_event).

    error: "all", "*", or a list of event names.
    warn: a list of event names.
    silence: a list of event names.
    valid_error_names: a set of event names that can be named in error, warn, and silence.

    In a hierarchy of configuration, the following rules apply:
    1. named > Deprecations > "all"/"*"
    2. silence > warn > error
    3. (1) > (2)
    """

    ERROR_ALL = ("all", "*")
    DEPRECATIONS = "Deprecations"

    error: Union[str, List[str]]
    warn: List[str]
    silence: List[str]

    def __init__(
        self,
        error: Optional[Union[str, List[str]]] = None,
        warn: Optional[List[str]] = None,
        silence: Optional[List[str]] = None,
        valid_error_names: Optional[Set[str]] = None,
    ):
        self._valid_error_names: Set[str] = valid_error_names or set()
        self._valid_error_names.add(self.DEPRECATIONS)

        # We can't do `= error or []` because if someone passes in an empty list, and latter appends to that list
        # they would expect references to the original list to be updated.
        self.error = error if error is not None else []
        self.warn = warn if warn is not None else []
        self.silence = silence if silence is not None else []

        # since we're overriding the dataclass auto __init__, we need to call __post_init__ manually
        self.__post_init__()

    def __post_init__(self):
        if isinstance(self.error, str) and self.error not in self.ERROR_ALL:
            raise ValidationError(f"error must be one of {self.ERROR_ALL} or a list of strings")

        # To specify `warn`, one of the following must be true
        # 1. `error` must be "all"/"*"
        # 2. "deprecations" must be in either `error` or `silence`.
        if self.warn and not (
            self.error in self.ERROR_ALL
            or self.DEPRECATIONS in self.error
            or self.DEPRECATIONS in self.silence
        ):
            raise ValidationError(
                f"`warn` can only be specified if `error` is one of {self.ERROR_ALL} or "
                f"{self.DEPRECATIONS} is in `error` or silence."
            )

        if isinstance(self.error, list):
            self._validate_items(self.error)

        if isinstance(self.warn, list):
            self._validate_items(self.warn)

        if isinstance(self.silence, list):
            self._validate_items(self.silence)

    def _validate_items(self, items: List[str]):
        for item in items:
            if item not in self._valid_error_names:
                raise ValidationError(f"{item} is not a valid dbt error name.")

    @property
    def _warn_error_options_v2(self) -> WarnErrorOptionsV2:
        # This is necessary because in core we directly set the WARN_ERROR_OPTIONS global variable
        # without this we'd need to do isinstance checks in `EventManager.warn_error_options`, which
        # would be costly as it gets called every time an event is fired.
        return self

    def _error_all(self) -> bool:
        """Is `*` or `all` set as error?"""
        return self.error in self.ERROR_ALL

    def _named_error(self, item_name: str) -> bool:
        """Is the item_name named in the error list?"""
        return item_name in self.error

    def _named_warn(self, item_name: str) -> bool:
        """Is the item_name named in the warn list?"""
        return item_name in self.warn

    def _named_silence(self, item_name: str) -> bool:
        """Is the item_name named in the silence list?"""
        return item_name in self.silence

    def _error_as_deprecation(self, event: Optional[BaseEvent]) -> bool:
        """Is the event a deprecation, and if so should it be treated as an error?"""
        return (
            event is not None and event.code().startswith("D") and self.DEPRECATIONS in self.error
        )

    def _warn_as_deprecation(self, event: Optional[BaseEvent]) -> bool:
        """Is the event a deprecation, and if so should it be treated as an warning?"""
        return (
            event is not None and event.code().startswith("D") and self.DEPRECATIONS in self.warn
        )

    def _silence_as_deprecation(self, event: Optional[BaseEvent]) -> bool:
        """Is the event a deprecation, and if so should it be silenced?"""
        return (
            event is not None
            and event.code().startswith("D")
            and self.DEPRECATIONS in self.silence
        )

    def errors(self, item_name: Union[str, BaseEvent]) -> bool:
        """Should the event be treated as an error?

        An event should error if any of the following are true:
        - The event is named in `error` and not named in `warn` or `silence`
        - "*" or "all" is specified for `error`, and the event is not named in `warn` or `silence`
        - The event is a deprecation, "deprecations" is in `error`, and the event is not named in `warn` or `silence`
          nor is "deprecations" in `warn` or `silence`
        """
        # Setup based on item_name type
        if isinstance(item_name, str):
            event_name = item_name
            event = None
        else:
            event_name = type(item_name).__name__
            event = item_name

        # Pre-compute checks that will be used multiple times
        named_elsewhere = self._named_warn(event_name) or self._named_silence(event_name)
        deprecation_elsewhere = self._warn_as_deprecation(event) or self._silence_as_deprecation(
            event
        )

        # Calculate result
        if self._named_error(event_name) and not named_elsewhere:
            return True
        elif self._error_as_deprecation(event) and not (named_elsewhere or deprecation_elsewhere):
            return True
        elif self._error_all() and not (named_elsewhere or deprecation_elsewhere):
            return True
        else:
            return False

    def includes(self, item_name: Union[str, BaseEvent]) -> bool:
        """Deprecated, use `errors` instead."""
        return self.errors(item_name)

    def silenced(self, item_name: Union[str, BaseEvent]) -> bool:
        """Is the event silenced?

        An event silenced if any of the following are true:
        - The event is named in `silence`
        - "Deprecations" is in `silence` and the event is not named in `error` or `warn`
        """
        # Setup based on item_name type
        if isinstance(item_name, str):
            event_name = item_name
            event = None
        else:
            event_name = type(item_name).__name__
            event = item_name

        # Pre-compute checks that will be used multiple times
        named_elsewhere = self._named_error(event_name) or self._named_warn(event_name)

        # Calculate result
        if self._named_silence(event_name):
            return True
        elif self._silence_as_deprecation(event) and not named_elsewhere:
            return True
        else:
            return False


FQNPath = Tuple[str, ...]
PathSet = AbstractSet[FQNPath]

T = TypeVar("T")


# A data type for representing lazily evaluated values.
#
# usage:
# x = Lazy.defer(lambda: expensive_fn())
# y = x.force()
#
# inspired by the purescript data type
# https://pursuit.purescript.org/packages/purescript-lazy/5.0.0/docs/Data.Lazy
@dataclass
class Lazy(Generic[T]):
    _f: Callable[[], T]
    memo: Optional[T] = None

    # constructor for lazy values
    @classmethod
    def defer(cls, f: Callable[[], T]) -> Lazy[T]:
        return Lazy(f)

    # workaround for open mypy issue:
    # https://github.com/python/mypy/issues/6910
    def _typed_eval_f(self) -> T:
        return cast(Callable[[], T], getattr(self, "_f"))()

    # evaluates the function if the value has not been memoized already
    def force(self) -> T:
        if self.memo is None:
            self.memo = self._typed_eval_f()
        return self.memo


# This class is used in to_target_dict, so that accesses to missing keys
# will return an empty string instead of Undefined
class DictDefaultEmptyStr(dict):
    def __getitem__(self, key):
        return dict.get(self, key, "")



from dbt_common.events.base_types import EventLevel
from dbt_common.events.event_manager_client import get_event_manager
from dbt_common.events.functions import get_stdout_config
from dbt_common.events.logger import LineFormat

# make sure event manager starts with a logger
get_event_manager().add_logger(
    get_stdout_config(LineFormat.PlainText, True, EventLevel.INFO, False)
)



from typing import List, Union, Any

from dbt_common.exceptions import CompilationError, CommandError, scrub_secrets, env_secrets


class SymbolicLinkError(CompilationError):
    def __init__(self) -> None:
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        msg = (
            "dbt encountered an error when attempting to create a symbolic link. "
            "If this error persists, please create an issue at: \n\n"
            "https://github.com/dbt-labs/dbt-core"
        )

        return msg


class ExecutableError(CommandError):
    def __init__(self, cwd: str, cmd: List[str], msg: str) -> None:
        super().__init__(cwd, cmd, msg)


class WorkingDirectoryError(CommandError):
    def __init__(self, cwd: str, cmd: List[str], msg: str) -> None:
        super().__init__(cwd, cmd, msg)

    def __str__(self, prefix: str = "! ") -> str:
        return f'{self.msg}: "{self.cwd}"'


class CommandResultError(CommandError):
    def __init__(
        self,
        cwd: str,
        cmd: List[str],
        returncode: Union[int, Any],
        stdout: bytes,
        stderr: bytes,
        msg: str = "Got a non-zero returncode",
    ) -> None:
        super().__init__(cwd, cmd, msg)
        self.returncode = returncode
        self.stdout = scrub_secrets(stdout.decode("utf-8"), env_secrets())
        self.stderr = scrub_secrets(stderr.decode("utf-8"), env_secrets())
        self.args = (cwd, self.cmd, returncode, self.stdout, self.stderr, msg)

    def __str__(self, prefix: str = "! ") -> str:
        return f"{self.msg} running: {self.cmd}"



import decimal
from _typeshed import Incomplete
from mashumaro.core.const import Sentinel
from typing import Any, Optional, Union
from typing_extensions import Literal

class SerializableType: ...
class GenericSerializableType: ...

class SerializationStrategy:
    def serialize(self, value: Any) -> Any: ...
    def deserialize(self, value: Any) -> Any: ...

class RoundedDecimal(SerializationStrategy):
    exp: Incomplete
    rounding: Incomplete
    def __init__(self, places: Optional[int] = ..., rounding: Optional[str] = ...) -> None: ...
    def serialize(self, value: decimal.Decimal) -> str: ...
    def deserialize(self, value: str) -> decimal.Decimal: ...

class Discriminator:
    field: Optional[str]
    include_supertypes: bool
    include_subtypes: bool
    def __post_init__(self) -> None: ...
    def __init__(self, field, include_supertypes, include_subtypes) -> None: ...



import os
import traceback
from typing import Any, List, Optional, Protocol, Tuple, Union

from dbt_common.events.base_types import BaseEvent, EventLevel, msg_from_base_event, TCallback
from dbt_common.events.logger import LoggerConfig, _Logger, _TextLogger, _JsonLogger, LineFormat
from dbt_common.exceptions.events import EventCompilationError
from dbt_common.helper_types import WarnErrorOptions, WarnErrorOptionsV2


class EventManager:
    def __init__(self) -> None:
        self.loggers: List[_Logger] = []
        self.callbacks: List[TCallback] = []
        self._warn_error: Optional[bool] = None
        self._warn_error_options: Optional[Union[WarnErrorOptions, WarnErrorOptionsV2]] = None
        self.require_warn_or_error_handling: bool = False

    @property
    def warn_error(self) -> bool:
        if self._warn_error is None:
            from dbt_common.events.functions import WARN_ERROR

            return WARN_ERROR
        return self._warn_error

    @warn_error.setter
    def warn_error(self, warn_error: bool) -> None:
        self._warn_error = warn_error

    @property
    def warn_error_options(self) -> Union[WarnErrorOptions, WarnErrorOptionsV2]:
        # Technically this always returns a WarnErrorOptionsV2, but to remain backwards compatible
        # with the protocol, we need to type the function as being able to return either.

        if self._warn_error_options is None:
            from dbt_common.events.functions import WARN_ERROR_OPTIONS

            return WARN_ERROR_OPTIONS._warn_error_options_v2

        return self._warn_error_options._warn_error_options_v2

    @warn_error_options.setter
    def warn_error_options(
        self, warn_error_options: Union[WarnErrorOptions, WarnErrorOptionsV2]
    ) -> None:
        self._warn_error_options = warn_error_options

    def fire_event(
        self,
        e: BaseEvent,
        level: Optional[EventLevel] = None,
        node: Any = None,
        force_warn_or_error_handling: bool = False,
    ) -> None:
        msg = msg_from_base_event(e, level=level)

        if force_warn_or_error_handling or (
            self.require_warn_or_error_handling and msg.info.level == "warn"
        ):
            if self.warn_error or self.warn_error_options.errors(e):
                # This has the potential to create an infinite loop if the handling of the raised
                # EventCompilationError fires an event as a warning instead of an error.
                raise EventCompilationError(e.message(), node)
            elif self.warn_error_options.silenced(e):
                # Return early if the event is silenced
                return

        if os.environ.get("DBT_TEST_BINARY_SERIALIZATION"):
            print(f"--- {msg.info.name}")
            try:
                msg.SerializeToString()
            except Exception as exc:
                raise Exception(
                    f"{msg.info.name} is not serializable to binary. ",
                    f"Originating exception: {exc}, {traceback.format_exc()}",
                )

        for logger in self.loggers:
            if logger.filter(msg):  # type: ignore
                logger.write_line(msg)

        for callback in self.callbacks:
            callback(msg)

    def add_logger(self, config: LoggerConfig) -> None:
        logger = (
            _JsonLogger(config) if config.line_format == LineFormat.Json else _TextLogger(config)
        )
        self.loggers.append(logger)

    def add_callback(self, callback: TCallback) -> None:
        self.callbacks.append(callback)

    def flush(self) -> None:
        for logger in self.loggers:
            logger.flush()


class IEventManager(Protocol):
    callbacks: List[TCallback]
    loggers: List[_Logger]
    warn_error: bool
    warn_error_options: Union[WarnErrorOptions, WarnErrorOptionsV2]
    require_warn_or_error_handling: bool

    def fire_event(
        self,
        e: BaseEvent,
        level: Optional[EventLevel] = None,
        node: Any = None,
        force_warn_or_error_handling: bool = False,
    ) -> None:
        ...

    def add_logger(self, config: LoggerConfig) -> None:
        ...

    def add_callback(self, callback: TCallback) -> None:
        ...


class TestEventManager(IEventManager):
    __test__ = False

    def __init__(self) -> None:
        self.event_history: List[Tuple[BaseEvent, Optional[EventLevel]]] = []
        self.loggers = []
        self.warn_error = False
        self.warn_error_options = WarnErrorOptions(include=[], exclude=[])
        self.require_warn_or_error_handling = False

    def fire_event(
        self,
        e: BaseEvent,
        level: Optional[EventLevel] = None,
        node: Any = None,
        force_warn_or_error_handling: bool = False,
    ) -> None:
        self.event_history.append((e, level))

    def add_logger(self, config: LoggerConfig) -> None:
        raise NotImplementedError()



import builtins
from typing import Any, Dict, List, Optional
import os

from dbt_common.constants import SECRET_ENV_PREFIX
from dbt_common.dataclass_schema import ValidationError


def env_secrets() -> List[str]:
    return [v for k, v in os.environ.items() if k.startswith(SECRET_ENV_PREFIX) and v.strip()]


def scrub_secrets(msg: Any, secrets: List[str]) -> Any:
    scrubbed = str(msg)

    for secret in secrets:
        scrubbed = scrubbed.replace(secret, "*****")

    return msg if str(msg) == scrubbed else scrubbed


class DbtBaseException(Exception):
    CODE = -32000
    MESSAGE = "Server Error"

    def data(self) -> Dict[str, Any]:
        # if overriding, make sure the result is json-serializable.
        return {
            "type": self.__class__.__name__,
            "message": str(self),
        }


class DbtInternalError(DbtBaseException):
    def __init__(self, msg: str) -> None:
        self.stack: List = []
        self.msg = scrub_secrets(msg, env_secrets())

    @property
    def type(self) -> str:
        return "Internal"

    def process_stack(self) -> List[str]:
        lines = []
        stack = self.stack
        first = True

        if len(stack) > 1:
            lines.append("")

            for item in stack:
                msg = "called by"

                if first:
                    msg = "in"
                    first = False

                lines.append(f"> {msg}")

        return lines

    def __str__(self) -> str:
        if hasattr(self.msg, "split"):
            split_msg = self.msg.split("\n")
        else:
            split_msg = str(self.msg).split("\n")

        lines = ["{}".format(self.type + " Error")] + split_msg

        lines += self.process_stack()

        return lines[0] + "\n" + "\n".join(["  " + line for line in lines[1:]])


class DbtRuntimeError(RuntimeError, DbtBaseException):
    CODE = 10001
    MESSAGE = "Runtime error"

    def __init__(self, msg: str, node=None) -> None:
        self.stack: List = []
        self.node = node
        self.msg = scrub_secrets(msg, env_secrets())

    def add_node(self, node=None) -> None:
        if node is not None and node is not self.node:
            if self.node is not None:
                self.stack.append(self.node)
            self.node = node

    @property
    def type(self):
        return "Runtime"

    def node_to_string(self, node: Any) -> str:
        """Given a node-like object we attempt to create the best identifier we can."""
        result = ""
        if hasattr(node, "resource_type"):
            result += node.resource_type
        if hasattr(node, "name"):
            result += f" {node.name}"
        if hasattr(node, "original_file_path"):
            result += f" ({node.original_file_path})"

        return result.strip() if result != "" else "<Unknown>"

    def process_stack(self) -> List[str]:
        lines = []
        stack = self.stack + [self.node]
        first = True

        if len(stack) > 1:
            lines.append("")

            for item in stack:
                msg = "called by"

                if first:
                    msg = "in"
                    first = False

                lines.append(f"> {msg} {self.node_to_string(item)}")

        return lines

    def validator_error_message(self, exc: builtins.Exception) -> str:
        """Given a dbt.dataclass_schema.ValidationError return the relevant parts as a string.

        dbt.dataclass_schema.ValidationError is basically a jsonschema.ValidationError)
        """
        if not isinstance(exc, ValidationError):
            return str(exc)
        path = "[%s]" % "][".join(map(repr, exc.relative_path))
        return f"at path {path}: {exc.message}"

    def __str__(self, prefix: str = "! ") -> str:
        node_string = ""

        if self.node is not None:
            node_string = f" in {self.node_to_string(self.node)}"

        if hasattr(self.msg, "split"):
            split_msg = self.msg.split("\n")
        else:
            split_msg = str(self.msg).split("\n")

        lines = ["{}{}".format(self.type + " Error", node_string)] + split_msg

        lines += self.process_stack()

        return lines[0] + "\n" + "\n".join(["  " + line for line in lines[1:]])

    def data(self) -> Dict[str, Any]:
        result = DbtBaseException.data(self)
        if self.node is None:
            return result

        result.update(
            {
                "raw_code": self.node.raw_code,
                # the node isn't always compiled, but if it is, include that!
                "compiled_code": getattr(self.node, "compiled_code", None),
            }
        )
        return result


class CompilationError(DbtRuntimeError):
    CODE = 10004
    MESSAGE = "Compilation Error"

    @property
    def type(self):
        return "Compilation"

    def _fix_dupe_msg(self, path_1: str, path_2: str, name: str, type_name: str) -> str:
        if path_1 == path_2:
            return (
                f"remove one of the {type_name} entries for {name} in this file:\n - {path_1!s}\n"
            )
        else:
            return (
                f"remove the {type_name} entry for {name} in one of these files:\n"
                f" - {path_1!s}\n{path_2!s}"
            )


class RecursionError(DbtRuntimeError):
    pass


class DbtConfigError(DbtRuntimeError):
    CODE = 10007
    MESSAGE = "DBT Configuration Error"

    # ToDo: Can we remove project?
    def __init__(self, msg: str, project=None, result_type="invalid_project", path=None) -> None:
        self.project = project
        super().__init__(msg)
        self.result_type = result_type
        self.path = path

    def __str__(self, prefix="! ") -> str:
        msg = super().__str__(prefix)
        if self.path is None:
            return msg
        else:
            return f"{msg}\n\nError encountered in {self.path}"


class NotImplementedError(DbtBaseException):
    def __init__(self, msg: str) -> None:
        self.msg = msg
        self.formatted_msg = f"ERROR: {self.msg}"
        super().__init__(self.formatted_msg)


class SemverError(Exception):
    def __init__(self, msg: Optional[str] = None) -> None:
        self.msg = msg
        if msg is not None:
            super().__init__(msg)
        else:
            super().__init__()


class VersionsNotCompatibleError(SemverError):
    pass


class DbtValidationError(DbtRuntimeError):
    CODE = 10005
    MESSAGE = "Validation Error"


class DbtDatabaseError(DbtRuntimeError):
    CODE = 10003
    MESSAGE = "Database Error"

    def process_stack(self) -> List[str]:
        lines = []

        if hasattr(self.node, "build_path") and self.node.build_path:
            lines.append(f"compiled code at {self.node.build_path}")

        return lines + DbtRuntimeError.process_stack(self)

    @property
    def type(self):
        return "Database"


class UnexpectedNullError(DbtDatabaseError):
    def __init__(self, field_name: str, source) -> None:
        self.field_name = field_name
        self.source = source
        msg = (
            f"Expected a non-null value when querying field '{self.field_name}' of table "
            f" {self.source} but received value 'null' instead"
        )
        super().__init__(msg)


class CommandError(DbtRuntimeError):
    def __init__(self, cwd: str, cmd: List[str], msg: str = "Error running command") -> None:
        cmd_scrubbed = list(scrub_secrets(cmd_txt, env_secrets()) for cmd_txt in cmd)
        super().__init__(msg)
        self.cwd = cwd
        self.cmd = cmd_scrubbed
        self.args = (cwd, cmd_scrubbed, msg)

    def __str__(self, prefix: str = "! ") -> str:
        if len(self.cmd) == 0:
            return f"{self.msg}: No arguments given"
        return f'{self.msg}: "{self.cmd[0]}"'



import datetime
from _typeshed import Incomplete

UTC_OFFSET_PATTERN: str

def parse_timezone(s: str) -> datetime.timezone: ...

class ConfigValue:
    name: Incomplete
    def __init__(self, name: str) -> None: ...



from _typeshed import Incomplete
from mashumaro.dialect import Dialect as Dialect
from mashumaro.helper import pass_through as pass_through
from mashumaro.mixins.dict import DataClassDictMixin as DataClassDictMixin
from typing import Any, Callable, Dict, TypeVar

T = TypeVar("T", bound="DataClassTOMLMixin")
EncodedData = str
Encoder = Callable[[Any], EncodedData]
Decoder = Callable[[EncodedData], Dict[Any, Any]]

class TOMLDialect(Dialect):
    omit_none: bool
    serialization_strategy: Incomplete

class DataClassTOMLMixin(DataClassDictMixin):
    def to_toml(self, encoder: Encoder = ..., **to_dict_kwargs: Any) -> EncodedData: ...
    @classmethod
    def from_toml(
        cls, data: EncodedData, decoder: Decoder = ..., **from_dict_kwargs: Any
    ) -> T: ...



import uuid
from datetime import datetime

_INVOCATION_ID = str(uuid.uuid4())
_INVOCATION_STARTED_AT = datetime.utcnow()


def get_invocation_id() -> str:
    return _INVOCATION_ID


def get_invocation_started_at() -> datetime:
    return _INVOCATION_STARTED_AT


def reset_invocation_id() -> None:
    global _INVOCATION_ID, _INVOCATION_STARTED_AT
    _INVOCATION_ID = str(uuid.uuid4())
    _INVOCATION_STARTED_AT = datetime.utcnow()



from dbt_common.exceptions import CompilationError, scrub_secrets, env_secrets


# event level exception
class EventCompilationError(CompilationError):
    def __init__(self, msg: str, node) -> None:
        self.msg = scrub_secrets(msg, env_secrets())
        self.node = node
        super().__init__(msg=self.msg)
